/*
 * $Id: rib.c,v 1.11 1998/07/17 18:36:24 masaki Exp $
 */

#include <stdlib.h>
#include <sys/types.h>
#include <ctype.h>
#include <time.h>

#include <mrt.h>
#include <config_file.h>
#include <stack.h>
#include <interface.h>
#include <rib.h>
#include <bgp.h>
#include <rip.h>
#include <ripng.h>
#include <array.h>

rib_t *RIB;
#ifdef HAVE_IPV6
rib_t *RIBv6;
#endif /* HAVE_IPV6 */

static u_int num_active_generic_attr = 0;
static u_int num_active_route_head = 0;
static u_int num_active_route_node = 0;
extern u_int num_active_prefixes;


static u_int 
rib_hash_fn (prefix_t * prefix, int size)
{
    u_int value;
    u_int buff[4];
    int len = 16;

    memset (buff, 0, sizeof (buff));
    if (prefix->family == AF_INET)
	len = 4;
    memcpy (buff, prefix_tochar (prefix), len);
    netmasking (prefix->family, (char *) buff, prefix->bitlen);

    value = buff[0] ^ buff[1] ^ buff[2] ^ buff[3];
    value ^= (value >> 16);
    value = value % size;
    return (value);
}


static u_int 
rib_lookup_fn (prefix_t * tmp1, prefix_t * tmp2)
{
    return (prefix_compare (tmp1, tmp2));
}


int
generic_attr_compare (generic_attr_t *a, generic_attr_t *b)
{
    return (a->type == b->type &&
	a->gateway == b->gateway &&
	  (a->nexthop == b->nexthop /* both are NULL */ ||
	  prefix_compare (a->nexthop, b->nexthop)) &&
	a->tag == b->tag);
}


char *
route_toa (prefix_t *prefix, generic_attr_t *attr, int pref)
{
    char *start, *buff;

    THREAD_SPECIFIC_STRAGE (buff);
    start = buff;
    
    if (attr->nexthop !=  NULL) 
      sprintf (buff, "%s nh %s",
	       prefix_toax (prefix), prefix_toa (attr->nexthop));
    else
      sprintf (buff, "%s nh unknown",   prefix_toax (prefix));
    buff = buff + strlen (buff);

    if (attr->gateway) {
	if (!prefix_compare (attr->gateway->prefix,
			     attr->nexthop)) {
	    sprintf (buff, " gw %s", prefix_toa (attr->gateway->prefix));
	    buff = buff + strlen (buff);
	}
    }
    sprintf (buff, " proto %s pref %d",
	     proto2string (attr->type), pref);
    return (start);
}


static route_node_t *
New_Route_Node (route_head_t *route_head,
		int pref, generic_attr_t *attr, u_long option)
{

    route_node_t *route_node = New (route_node_t);
    route_node->route_head = route_head;
    route_node->pref = pref;
    route_node->option = option;
    route_node->attr = Ref_Generic_Attr (attr);
    time (&route_node->time);
    num_active_route_node++;
    
    return (route_node);
}


static void 
Delete_Route_Node (route_node_t * rt_node)
{
    Deref_Generic_Attr (rt_node->attr);
    Delete (rt_node);
    num_active_route_node--;
}


static route_head_t *
New_Route_Head (prefix_t * prefix)
{
    route_head_t *route_head = New (route_head_t);

    route_head->prefix = Ref_Prefix (prefix);
    route_head->ll_route_nodes = 
	LL_Create (LL_DestroyFunction, Delete_Route_Node, 0);
    route_head->active = NULL;

    num_active_route_head++;
    return (route_head);
}


static void 
Delete_Route_Head (route_head_t *route_head)
{

    Deref_Prefix (route_head->prefix);
    LL_Destroy (route_head->ll_route_nodes);
    Delete (route_head);
    num_active_route_head--;
}


generic_attr_t *
New_Generic_Attr (int proto)
{
    generic_attr_t *attr;

    attr = New (generic_attr_t);
    attr->type = proto;
    attr->gateway = NULL;
    attr->nexthop = NULL;
    attr->ref_count = 1;
    pthread_mutex_init (&attr->mutex_lock, NULL);

    num_active_generic_attr++;
    return (attr);
}


generic_attr_t *
Ref_Generic_Attr (generic_attr_t * attr)
{
    pthread_mutex_lock (&attr->mutex_lock);
    attr->ref_count++;
    assert (attr->ref_count > 0);
    pthread_mutex_unlock (&attr->mutex_lock);
    return (attr);
}


void 
Deref_Generic_Attr (generic_attr_t * attr)
{
#ifdef notdef
    if (attr->type == PROTO_RIPNG)
        ripng_deref_attr ((ripng_attr_t *)attr);
    else if (attr->type == PROTO_BGP)
	bgp_deref_attr ((bgp_attr_t *)attr);
    else {
#else
    {
#endif
        pthread_mutex_lock (&attr->mutex_lock);
        if (--attr->ref_count <= 0) {
	    Deref_Prefix (attr->nexthop);
            pthread_mutex_destroy (&attr->mutex_lock);
	    Delete (attr);
    	    num_active_generic_attr--;
	    return;
	}
        pthread_mutex_unlock (&attr->mutex_lock);
    }
}


rib_t *
New_Rib (int maxbitlen)
{
    route_head_t head;
    rib_t *rib = New (rib_t);

    /*rib->ll_route_nodes = LL_Create (0); */
    rib->ll_changed_route_nodes = LL_Create (0);

    /*if (LL_ATTR_MEMORY == NULL)
       init_attr_memory (); */

    if (pthread_mutex_init (&rib->mutex_lock, NULL) != 0) {
	perror ("rib pthread_mutex_init");
	exit (0);
    }
    rib->hash = HASH_Create (100,
			  HASH_KeyOffset, HASH_Offset (&head, &head.prefix),
			     HASH_LookupFunction, rib_lookup_fn,
			     HASH_HashFunction, rib_hash_fn,
#if 0
			     HASH_ReportChange, True,
			     HASH_ReportAccess, True,
#endif
			     NULL);

    return (rib);
}


void
init_rib (trace_t * tr)
{

    assert (RIB == NULL);
    RIB = New_Rib (32);
    RIB->trace = trace_copy (tr);
    set_trace (RIB->trace, TRACE_PREPEND_STRING, "RIB", 0);

#ifdef HAVE_IPV6
    assert (RIBv6 == NULL);
    RIBv6 = New_Rib (128);
    RIBv6->trace = trace_copy (tr);
    set_trace (RIBv6->trace, TRACE_PREPEND_STRING, "RIB6", 0);
#endif /* HAVE_IPV6 */
}


static void
rib_kernel_update_route (rib_t *rib, u_int cmd, prefix_t * dest, 
			 generic_attr_t * new, generic_attr_t * old)
{
    int index = 0, oldindex = 0, rc;
    prefix_t *nexthop = NULL, *oldhop = NULL;
    int (*fn)();

    if ((fn = rib->kernel_update_route) == NULL)
	return;

    if (new) {
	assert (new->gateway);
        index = (new->gateway->interface)?new->gateway->interface->index:0;
        nexthop = new->nexthop;
    }
    if (old) {
	assert (old->gateway);
        oldindex = (old->gateway->interface)?old->gateway->interface->index:0;
        oldhop = old->nexthop;
    }
    rc = (*fn) (cmd, dest, nexthop, oldhop, index, oldindex);

    trace (TR_TRACE, rib->trace, "KERNEL %s %s nexthop %s index %d %s\n",
	   (cmd == KERNEL_ROUTE_ADD) ? "ADD" : 
	   (cmd == KERNEL_ROUTE_DEL) ? "DEL" :
	   (cmd == KERNEL_ROUTE_CHG) ? "CHG" : "???",
	   prefix_toax (dest),
	   prefix_toa (nexthop), index,
	   (rc < 0) ? "FAILED" : "OK");
}


/* rib_get_route_head
 * get "head" of routing table entry for a given prefix. Create a head
 * and radix-tree entry if one does not aleady exist
 */
static route_head_t *
rib_get_route_head (rib_t * rib, prefix_t * prefix)
{
    route_head_t *route_head;

    assert (pthread_mutex_trylock (&rib->mutex_lock));

    route_head = HASH_Lookup (rib->hash, prefix);

    if (route_head == NULL) {
	route_head = New_Route_Head (prefix);
	HASH_Insert (rib->hash, route_head);
	trace (TR_TRACE, rib->trace, "new route head for %s\n",
	       prefix_toax (prefix));
    }

    return (route_head);
}


/* rib_find_route_node
 * given a prefix AND type, find route node entry
 * return NULL if one does not exist
 */
static route_node_t *
rib_find_route_node (rib_t * rib,
		     prefix_t * prefix, generic_attr_t * attr)
{
    route_head_t *route_head;
    route_node_t *route_node;

    assert (pthread_mutex_trylock (&rib->mutex_lock));

    if ((route_head = HASH_Lookup (rib->hash, prefix)) != NULL) {
	LL_Iterate (route_head->ll_route_nodes, route_node) {
	    if (route_node->attr->type == attr->type)
		return (route_node);
	}
    }

    return (NULL);
}


static void 
rib_inter_proto_update (rib_t *rib, prefix_t * prefix, generic_attr_t * new,
			generic_attr_t * old, int pref, int from)
{
    int bgp = 0, ripng = 0, rip = 0; /* have been sent */
    proto_update_route_t fn;

    if (/* from == PROTO_BGP || */
	    !BIT_TEST (MRT->protocols, (1 << PROTO_BGP)))
	bgp++;
    if (prefix->family != AF_INET ||
	    /* from == PROTO_RIP || */
	    !BIT_TEST (MRT->protocols, (1 << PROTO_RIP)))
	rip++;
#ifdef HAVE_IPV6
    if (prefix->family != AF_INET6 ||
            /* from == PROTO_RIPNG || */
	    !BIT_TEST (MRT->protocols, (1 << PROTO_RIPNG)))
	ripng++;
#endif /* HAVE_IPV6 */

    /* to avoid loop */
    /* Redistribute should be processed by each protocol routine
       to be overridden by the active route */

    if (!bgp && new && old && new->type != PROTO_BGP && old->type != PROTO_BGP) {
/*        trace (TR_TRACE, rib->trace, "UPD %s -> BGP %s\n",
	    proto2string (old->type), prefix_toax (prefix)); */
	/* bgp routine can overwrite it */
	if ((fn = rib->proto_update_route[PROTO_BGP]) != NULL)
	    (*fn) (prefix, new, NULL, pref);
        bgp++;
    }
    if (!rip && new && old && new->type != PROTO_RIP && old->type != PROTO_RIP) {
/*        trace (TR_TRACE, rib->trace, "DEL %s -> RIP %s\n",
	    proto2string (old->type), prefix_toax (prefix));
        trace (TR_TRACE, rib->trace, "ADD %s -> RIP %s\n",
	    proto2string (new->type), prefix_toax (prefix)); */
	/* XXX I don't know how to update it */
	if ((fn = rib->proto_update_route[PROTO_RIP]) != NULL)
	    (*fn) (prefix, new, old, pref);
    }
#ifdef HAVE_IPV6
    if (!ripng && new && old && new->type != PROTO_RIPNG && 
			   old->type != PROTO_RIPNG) {
/*        trace (TR_TRACE, rib->trace, "UPD %s -> RIPNG %s\n",
	    proto2string (new->type), prefix_toax (prefix)); */
	/* ripng routine can always overwrite it */
	if ((fn = rib->proto_update_route[PROTO_RIPNG]) != NULL)
	    (*fn) (prefix, new, NULL, pref);
	ripng++;
    }
#endif /* HAVE_IPV6 */

    if (old) {
	if (!bgp && old->type != PROTO_BGP) {
/*            trace (TR_TRACE, rib->trace, "DEL %s -> BGP %s\n",
	        proto2string (old->type), prefix_toax (prefix)); */
	if ((fn = rib->proto_update_route[PROTO_BGP]) != NULL)
	    (*fn) (prefix, NULL, old, 0);
	}

	if (!rip && old->type != PROTO_RIP) {
/*            trace (TR_TRACE, rib->trace, "DEL %s -> RIP %s\n",
	        proto2string (old->type), prefix_toax (prefix)); */
	    /* XXX I don't know how to delete it */
	    if ((fn = rib->proto_update_route[PROTO_RIP]) != NULL)
	        (*fn) (prefix, NULL, old, 0);
	}
#ifdef HAVE_IPV6
	if (!ripng && old->type != PROTO_RIPNG) {
/*            trace (TR_TRACE, rib->trace, "DEL %s -> RIPNG %s\n",
	        proto2string (old->type), prefix_toax (prefix)); */
	    if ((fn = rib->proto_update_route[PROTO_RIPNG]) != NULL)
	        (*fn) (prefix, NULL, old, 0);
	}
#endif /* HAVE_IPV6 */
    }

    if (new) {
	if (!bgp && new->type != PROTO_BGP) {
/*            trace (TR_TRACE, rib->trace, "ADD %s -> BGP %s\n",
	        proto2string (new->type), prefix_toax (prefix)); */
	    if ((fn = rib->proto_update_route[PROTO_BGP]) != NULL)
	        (*fn) (prefix, new, NULL, pref);
	}

	if (!rip && new->type != PROTO_RIP) {
/*            trace (TR_TRACE, rib->trace, "ADD %s -> RIP %s\n",
	        proto2string (new->type), prefix_toax (prefix)); */
	    if ((fn = rib->proto_update_route[PROTO_RIP]) != NULL)
	        (*fn) (prefix, new, NULL, pref);
#ifdef notdef
	    rip_attr_t *attr = New_Rip_Attr (1);

	    assert (new->type != PROTO_RIP);
	    attr->type = new->type;
	    attr->nexthop = Ref_Prefix (new->nexthop);
	    attr->gateway = new->gateway;
	    attr->metric = 1;
	    attr->tag = 0;
	    attr->ref_count = 1;	/* should be done in New_Rip_Attr */
	    pthread_mutex_init (&attr->mutex_lock, NULL);
	    New_Rip_Route (prefix, attr);
#endif
	}
#ifdef HAVE_IPV6
	if (!ripng && new->type != PROTO_RIPNG) {
/*            trace (TR_TRACE, rib->trace, "ADD %s -> RIPNG %s\n",
	        proto2string (new->type), prefix_toax (prefix)); */
	    if ((fn = rib->proto_update_route[PROTO_RIPNG]) != NULL)
	        (*fn) (prefix, new, NULL, pref);
	}
#endif /* HAVE_IPV6 */
    }
}


static void 
rib_replace_active (rib_t * rib, route_head_t * route_head,
		    route_node_t * best, int from)
{
    route_node_t *route_node;

    if (best == NULL) {
	/* find new active route */
	LL_Iterate (route_head->ll_route_nodes, route_node) {
	    if (best == NULL || route_node->pref < best->pref) {
		best = route_node;
	    }
	}
    }

    if (best == NULL) {	/* there is no other route nodes */

	assert (route_head->active);
	rib_kernel_update_route (rib, KERNEL_ROUTE_DEL, route_head->prefix,
				 route_head->active->attr, NULL);
	trace (TR_TRACE, rib->trace, "delete route head for %s\n",
	       prefix_toax (route_head->prefix));
	if (!BIT_TEST (route_head->active->option, MRT_RTOPT_SUPPRESS))
            rib_inter_proto_update (rib, route_head->prefix, NULL, 
				    route_head->active->attr, 0, from);
	HASH_RemoveFn (rib->hash, route_head, 0); 
	Delete_Route_Head (route_head);
	return;
    }

    if (route_head->active == best)
	return;

    /* route update */

    trace (TR_TRACE, rib->trace, "active: %s%s\n",
	   route_toa (route_head->prefix, best->attr, best->pref),
	   BIT_TEST (best->option, MRT_RTOPT_SUPPRESS)? " (suppressed)": "");

    if (route_head->active == NULL) {
       if (!BIT_TEST (best->option, MRT_RTOPT_SUPPRESS))
            rib_inter_proto_update (rib, route_head->prefix, best->attr, 
	        NULL, best->pref, from);
    }
    else {
        if (BIT_TEST (best->option, MRT_RTOPT_SUPPRESS) &&
	    BIT_TEST (route_head->active->option, MRT_RTOPT_SUPPRESS)) {
	    /* do nothing */
        }
        else if (BIT_TEST (best->option, MRT_RTOPT_SUPPRESS) &&
            !BIT_TEST (route_head->active->option, MRT_RTOPT_SUPPRESS)) {
            rib_inter_proto_update (rib, route_head->prefix, NULL,
	            route_head->active->attr, 0, from);
        }
        else if (!BIT_TEST (best->option, MRT_RTOPT_SUPPRESS) &&
            BIT_TEST (route_head->active->option, MRT_RTOPT_SUPPRESS)) {
            rib_inter_proto_update (rib, route_head->prefix, best->attr, 
	            NULL, best->pref, from);
        }
        else if (!BIT_TEST (best->option, MRT_RTOPT_SUPPRESS) &&
            !BIT_TEST (route_head->active->option, MRT_RTOPT_SUPPRESS)) {
            rib_inter_proto_update (rib, route_head->prefix, best->attr, 
	            route_head->active->attr, best->pref, from);
        }
    }

    if (route_head->active &&
	prefix_compare (route_head->active->attr->nexthop,
			best->attr->nexthop) /* nexthop unchange */) {
	/* they are the same, so it's OK if replaced. */
	/* this is required in case active is being deleted. */
	route_head->active = best;
	return;
    }

    if (BIT_TEST (best->option, MRT_RTOPT_NOINSTALL) && route_head->active) {
	rib_kernel_update_route (rib, KERNEL_ROUTE_DEL, route_head->prefix,
				 route_head->active->attr, NULL);
        route_head->active = best;
	return;
    }
    if (best->attr->type != PROTO_KERNEL &&
	best->attr->type != PROTO_CONNECTED &&
	!BIT_TEST (best->option, MRT_RTOPT_NOINSTALL)) {

	if (route_head->active &&
	    !BIT_TEST (route_head->active->option, MRT_RTOPT_NOINSTALL)) {
	    rib_kernel_update_route (rib, KERNEL_ROUTE_CHG, 
			route_head->prefix, best->attr, 
			route_head->active->attr);
	}
	else {
	    rib_kernel_update_route (rib, KERNEL_ROUTE_ADD,
			route_head->prefix, best->attr, NULL);
	}
    }
    route_head->active = best;
}


static void 
rib_delete_route_node (rib_t * rib, route_node_t * route_node)
{
    route_head_t *route_head;

    assert (pthread_mutex_trylock (&rib->mutex_lock));

    route_head = route_node->route_head;
    LL_RemoveFn (route_head->ll_route_nodes, route_node, 0);

    trace (TR_TRACE, rib->trace, "delete: %s\n",
	route_toa (route_head->prefix, route_node->attr, route_node->pref));

    if (route_node->protocol_mask == 0) {
	/* what's protocol_mask? */
    }

    if (route_head->active == route_node) {
	rib_replace_active (rib, route_head, NULL, route_node->attr->type);
    }

    /* delete here */
    Delete_Route_Node (route_node);
}



int 
rib_route_sweep (rib_t * rib, int type)
{
    route_head_t *route_head;
    route_node_t *route_node;
    LINKED_LIST *ll;

    assert (pthread_mutex_trylock (&rib->mutex_lock));

    ll = LL_Create (0);
    HASH_Iterate (rib->hash, route_head) {
	LL_Iterate (route_head->ll_route_nodes, route_node) {
	    if (route_node->attr->type == type)
		LL_Add (ll, route_node);
	}
    }
    LL_Iterate (ll, route_node) {
	rib_delete_route_node (rib, route_node);
    }
    LL_Destroy (ll);
    return (1);
}


int 
show_rib_status (uii_connection_t *uii)
{
    uii_send_data (uii, "%d active prefixes\n", num_active_prefixes);
    uii_send_data (uii, "%d active generic attributes\n", 
		num_active_generic_attr);
    uii_send_data (uii, "%d active route heads\n", num_active_route_head);
    uii_send_data (uii, "%d active route nodes\n", num_active_route_node);
    show_gateway (uii);
    return (1);
}


static int
rib_route_head_compare (route_head_t *a, route_head_t *b)
{
    return (prefix_compare2 (a->prefix, b->prefix));
}


void 
rib_show_route_head (uii_connection_t *uii, char *append)
{
    uii_send_data (uii, "%c%c%c %-4s %-8s %-26s %-24s %-4s %s\n",
	' ', ' ', 'P', "Pref", "Time", "Destination", "Next Hop", "If", 
	(append)? append: "");
}


void 
rib_show_route_line (uii_connection_t *uii, int c1, int c2, int type, int pref,
        time_t time, prefix_t *prefix, prefix_t *nexthop, gateway_t *gateway, 
	char *append)
{
    char tmp1[MAXLINE], date[MAXLINE];;

    sprintf (tmp1, "%s", prefix_toax (prefix));
    if (time/3600 > 99)
	sprintf (date, "%02lddy%02ldhr", time/(3600*24), (time%(3600*24))/3600);
    else
        sprintf (date, "%02ld:%02ld:%02ld", time/3600, (time/60)%60, time%60);
    
    if (nexthop != NULL) 
      uii_send_data (uii, "%c%c%c %4d %-8s %-26s %-24s %-4s %s\n", 
		     c1, c2, toupper (proto2string (type)[0]),
		     pref, date, tmp1,
		     prefix_toa (nexthop),
		     (gateway && gateway->interface) ?
		     gateway->interface->name : "",
		     (append)? append: "");
    else
      uii_send_data (uii, "%c%c%c %4d %-8s %-26s %-24s %-4s %s\n", 
		     c1, c2, toupper (proto2string (type)[0]),
		     pref, date, tmp1,
		     "unknown", 
		     (gateway && gateway->interface) ?
		     gateway->interface->name : "",
		     (append)? append: "");
}


int 
show_rib_routes (uii_connection_t *uii, rib_t *rib)
{
    route_head_t *route_head;
    route_node_t *route_node;
    DATA_PTR *array;
    int i;
    u_int nel;
    time_t now;

    rib_open (rib);
    array = HASH_ToArray (rib->hash, NULL, &nel);
    rib_close (rib);
    ARRAY_Sort (array, nel, rib_route_head_compare);

    time (&now);
    uii_send_data (uii, "%d prefixes\n", nel);
    rib_show_route_head (uii, NULL);

    for (i = 0; i < nel; i++) {
        route_head = (route_head_t *) array[i];

	LL_Iterate (route_head->ll_route_nodes, route_node) {
	    int c = (route_head->active == route_node) ? '*' : ' ';
            time_t time = now - route_node->time;

	    rib_show_route_line (uii, c, ' ', route_node->attr->type,
		route_node->pref, time, 
		route_head->prefix, route_node->attr->nexthop,
		route_node->attr->gateway, NULL);
	}
    }
    Delete (array);
    return (1);
}



/* 
 * 0 = add, 1 = change, 2= delete 
 */
int 
rib_update_route (int cmd, prefix_t *prefix, generic_attr_t *attr, 
		  int pref, u_long option)
{
    rib_t *rib;

    if (prefix->family == AF_INET)
        rib = RIB;
#ifdef HAVE_IPV6
    else if (prefix->family == AF_INET6)
        rib = RIBv6;
#endif /* HAVE_IPV6 */
    else
        return (-1);
    
    rib_open (rib);
    switch (cmd) {
    case RIB_UPD:
	rib_add_route (rib, prefix, attr, pref, option);
	break;
    case RIB_DEL:
	rib_del_route (rib, prefix, attr);
	break;
#ifdef notdef
    case RIB_HOLD:
	rib_hold_route (rib, prefix, attr, HOLD_PREF);
	break;
    case RIB_UNHOLD:
	rib_hold_route (rib, prefix, attr, pref);
	break;
#endif
    default:
	assert (0);
	break;
    }
    rib_close (rib);
    return (1);
}


/* 
 * add a route to the appropriate rib
 */
int 
add_route_to_rib (int type, prefix_t *prefix, prefix_t *nexthop, 
		  interface_t *interface, int pref, u_long opt)
{
    generic_attr_t *attr;
    rib_t *rib;
    int ret = 1;

    assert (prefix);
    assert (nexthop);
    assert (prefix->family == nexthop->family);

    rib = RIB;
#ifdef HAVE_IPV6
    if (prefix->family == AF_INET6)
        rib = RIBv6;
#endif /* HAVE_IPV6 */

    attr = New_Generic_Attr (type);
    attr->nexthop = Ref_Prefix (nexthop);
    attr->gateway = add_gateway (attr->nexthop, 0, interface);
    rib_open (rib);
    if (rib_add_route (rib, prefix, attr, pref, opt) == NULL)
	ret = -1;
    rib_close (rib);
    Deref_Generic_Attr (attr);
    return (ret);
}


/* 
 * delete a route from the appropriate rib
 */
int 
del_route_from_rib (int type, prefix_t *prefix)
{
    generic_attr_t *attr;
    rib_t *rib;
    int ret;

    assert (prefix);

    rib = RIB;
#ifdef HAVE_IPV6
    if (prefix->family == AF_INET6)
        rib = RIBv6;
#endif /* HAVE_IPV6 */

    attr = New_Generic_Attr (type);
    attr->nexthop = NULL;
    attr->gateway = NULL;
    rib_open (rib);
    ret = rib_del_route (rib, prefix, attr);
    rib_close (rib);
    Deref_Generic_Attr (attr);
    return (ret);
}


void
delete_all_route_with_type (int family, int type)
{
    if (family == AF_INET) {
        rib_open (RIB);
        rib_route_sweep (RIB, type);
        rib_close (RIB);
    }
#ifdef HAVE_IPV6
    if (family == AF_INET6) {
        rib_open (RIBv6);
        rib_route_sweep (RIBv6, type);
        rib_close (RIBv6);
    }
#endif /* HAVE_IPV6 */
}


int
show_ip_routes (uii_connection_t * uii)
{
    return (show_rib_routes (uii, RIB));
}


#ifdef HAVE_IPV6
int
show_ipv6_routes (uii_connection_t * uii)
{
    return (show_rib_routes (uii, RIBv6));
}
#endif /* HAVE_IPV6 */


static int 
trace_bgp_rib_op (uii_connection_t * uii, char *s, int op)
{
    if (strcasecmp (s, "*") == 0) {
        rib_open (RIB);
       	set_trace (RIB->trace, op, TR_ALL, NULL);
        rib_close (RIB);
#ifdef HAVE_IPV6
        rib_open (RIBv6);
       	set_trace (RIBv6->trace, op, TR_ALL, NULL);
        rib_close (RIBv6);
#endif /* HAVE_IPV6 */
	
    }
    else if (strcasecmp (s, "inet") == 0) {
        rib_open (RIB);
       	set_trace (RIB->trace, op, TR_ALL, NULL);
        rib_close (RIB);
    }
#ifdef HAVE_IPV6
    else if (strcasecmp (s, "inet6") == 0) {
        rib_open (RIBv6);
       	set_trace (RIBv6->trace, op, TR_ALL, NULL);
        rib_close (RIBv6);
    }
#endif /* HAVE_IPV6 */
    else {
	config_notice (TR_ERROR, uii, 
		       "invalid or unconfigured rib %s\n", s);
        Delete (s);
        return (-1);
    }
    Delete (s);
    return (1);
}


int 
trace_rib (uii_connection_t * uii, char *s)
{
    return (trace_bgp_rib_op (uii, s, TRACE_ADD_FLAGS));
}


int 
no_trace_rib (uii_connection_t * uii, char *s)
{
    return (trace_bgp_rib_op (uii, s, TRACE_DEL_FLAGS));
}


/* rib_add_route
 * create a route_node and add it to the appropriate route_head,
 * or create a route_head if not already exist. Returnthe new route_node.
 */

route_node_t *
rib_add_route (rib_t *rib, prefix_t *prefix, generic_attr_t *attr, 
	       int pref, u_long option)
{
    route_node_t *route_node;
    route_head_t *route_head;

    assert (pthread_mutex_trylock (&rib->mutex_lock));

    route_head = rib_get_route_head (rib, prefix);

    if (attr->type == PROTO_KERNEL) {
	/* KERNEL route servives only when it's active */
        if (route_head->active) {
	    if (route_head->active->attr->type == PROTO_KERNEL) {
		/* delete this 
		   Linux has multiple routes for the same destination */
	        trace (TR_TRACE, rib->trace,
		   "duplicate KERNEL route: %s (deleting)\n",
	          route_toa (prefix, attr, pref));
		rib_kernel_update_route (rib, KERNEL_ROUTE_DEL, prefix, 
					 attr, NULL);
	    }
	    else {
		/* ignore this */
	        trace (TR_ERROR, rib->trace,
		   "KERNEL route: %s shouldn't exist (ignored)\n",
	          route_toa (prefix, attr, pref));
	    }
	    return (NULL);
	}
        /* no active route */
        route_node = New_Route_Node (route_head, pref, attr, option);
        LL_Add (route_head->ll_route_nodes, route_node);
	rib_replace_active (rib, route_head, route_node, attr->type);
	return (route_node);
    }

    /* check to see if we already got this route from the same proto */

    LL_Iterate (route_head->ll_route_nodes, route_node) {
        if (/* a route from the kernel always matches */
	        route_node->attr->type == PROTO_KERNEL ||
		route_node->attr->type == attr->type)
	    break;
    }

    if (route_node) {
	route_node_t *new_route;

	if (attr->type == route_node->attr->type 
	    /* same proto is always overriden */ ||
	    pref < route_node->pref) {
	    /* better preference (only against kernel route) */

	    if (pref == route_node->pref &&
		    generic_attr_compare (attr, route_node->attr) &&
		    option == route_node->option) {
	        trace (TR_TRACE, rib->trace,
		   "SAME: %s (ignored)\n",
	          route_toa (prefix, route_node->attr, route_node->pref));
	        return (route_node);
	    }

	    /* implicit withdarw of old route -- delete it */
	    /* on deleting, 
		do I have to notify the proto engine which supplied? */

	    trace (TR_TRACE, rib->trace,
		   "update: %s -> %d\n",
	      route_toa (prefix, route_node->attr, route_node->pref), pref);

    	    new_route = New_Route_Node (route_head, pref, attr, option);
    	    LL_Add (route_head->ll_route_nodes, new_route);
	    LL_RemoveFn (route_head->ll_route_nodes, route_node, 0);

	    rib_replace_active (rib, route_head, NULL, attr->type);

    	    Delete_Route_Node (route_node);
	    return (route_node);
	}
	else {
	    /* same or bad .. ignore it */

	    trace (TR_TRACE, rib->trace,
		   "ignore: %s (now %d)\n",
		   route_toa (prefix, attr, pref), route_node->pref);

	    return (route_node);
	}
    }

    /* first time we've heard from this gateway in this proto */
    route_node = New_Route_Node (route_head, pref, attr, option);
    LL_Add (route_head->ll_route_nodes, route_node);

    trace (TR_TRACE, rib->trace,
	   "add: %s\n", route_toa (prefix, attr, pref));
    /* is this new route better than the active route ? */
    if (route_head->active == NULL ||
	route_node->pref < route_head->active->pref) {
	rib_replace_active (rib, route_head, route_node, attr->type);
    }

    return (route_node);
}

/* XXX
 * attr is not used seriously, only attr->type is unsed.
 */
int
rib_del_route (rib_t * rib, prefix_t * prefix, generic_attr_t * attr)
{
    route_node_t *route_node;

    assert (pthread_mutex_trylock (&rib->mutex_lock));
    if ((route_node = rib_find_route_node (rib, prefix, attr)) == NULL)
	return (-1);
    assert (route_node);
    rib_delete_route_node (rib, route_node);
    return (1);
}


void 
rib_open (rib_t * rib)
{
    if (pthread_mutex_trylock (&rib->mutex_lock) != 0) {
	trace (TR_DEBUG, rib->trace, "Going to block in rib_open\n");
	pthread_mutex_lock (&rib->mutex_lock);
    }
}


void 
rib_close (rib_t * rib)
{
    pthread_mutex_unlock (&rib->mutex_lock);
}


/* called from BGP to look up a nexthop for mutihop neighbor */
prefix_t *
rib_find_best_route (prefix_t *prefix)
{
    rib_t *rib = RIB;
    route_head_t *route_head;
    prefix_t *nexthop = NULL;
    int maxlen = 32, save, i;

#ifdef HAVE_IPV6
    if (prefix->family == AF_INET6) {
	rib = RIBv6;
	maxlen = 128;
    }
#endif /* HAVE_IPV6 */
    rib_open (rib);
    /* I know this is a very bad way of finding the best route */
    save = prefix->bitlen;
    for (i = maxlen; i >= 0; i--) {
	prefix->bitlen = i;
        route_head = HASH_Lookup (rib->hash, prefix);
        if (route_head && route_head->active) {
	    nexthop = route_head->active->attr->nexthop;
	    break;
        }
    }
    prefix->bitlen = save;
    rib_close (rib);
    return (nexthop);
}
